import re
from pathlib import Path
from typing import Any, Dict, List, Optional

from freeds.config import get_config


class PluginError(Exception):
    pass


def get_repo_dir(repo: str) -> Path:
    start = Path.cwd()
    while True:
        if start.name == repo:
            return start
        if (start / repo).exists():
            return start / repo
        if start.parent is None:
            return None
        start = start.parent


def is_ip(s: str) -> bool:
    # Determine wheter a string is an ip address, accepts IPv4 or IPv6 (in brackets).
    if s.startswith("[") and s.endswith("]"):
        return True
    if s.count(".") == 3 and all(p.isdigit() and 0 <= int(p) <= 255 for p in s.split(".")):
        return True
    return False


class PortMapping:
    """
    Parses Docker Compose port mapping strings.
    Examples:
        127.0.0.1:1234:1234
        1234:1234
        127.0.0.1:1234:0.0.0.0:4321
        1234:0.0.0.0:1234
        127.0.0.1:1234:0.0.0.1:1234/udp
        [::1]:1234:1234
        1234:[::1]:1234
        [::1]:1234:[::2]:4321/udp
    """

    def __init__(self, mapping: str) -> None:
        self.mapping = mapping
        self.host_ip: Optional[str] = None
        self.host_port: int = -1
        self.container_ip: Optional[str] = None
        self.container_port: int = -1
        self.protocol: Optional[str] = None
        self._parse()

    def _parse(self) -> None:
        # Accepts IPv4 and IPv6 addresses in port mapping strings
        # Handles:
        #   [::1]:1234:1234
        #   1234:[::1]:1234
        #   [::1]:1234:[::2]:4321/udp
        mapping = self.mapping
        if "/" in mapping:
            mapping, self.protocol = mapping.rsplit("/", 1)
        # Replace IPv6 colons with a placeholder to avoid splitting inside IPv6
        mapping = re.sub(r"(\[[^\]]+\])", lambda m: m.group(1).replace(":", "#"), mapping)
        parts = mapping.split(":")
        # Restore colons in IPv6 addresses
        parts = [p.replace("#", ":") for p in parts]

        if len(parts) == 2:
            self.host_port = int(parts[0])
            self.container_port = int(parts[1])
        elif len(parts) == 3:
            if is_ip(parts[1]):
                self.host_port = int(parts[0])
                self.container_ip = parts[1]
                self.container_port = int(parts[2])
            else:
                self.host_ip = parts[0]
                self.host_port = int(parts[1])
                self.container_port = int(parts[2])
        elif len(parts) == 4:
            self.host_ip = parts[0]
            self.host_port = int(parts[1])
            self.container_ip = parts[2]
            self.container_port = int(parts[3])
        else:
            raise ValueError(f"Invalid port mapping format: {self.mapping}")

    def __str__(self) -> str:
        return f"{port.host_ip}:{port.host_port} -> {port.container_ip}:{port.container_port} ({port.protocol})"

    def __repr__(self) -> str:
        return (
            f"PortMapping(host_ip={self.host_ip}, host_port={self.host_port}, "
            f"container_ip={self.container_ip}, container_port={self.container_port}, protocol={self.protocol})"
        )


class Repo:
    def __init__(self, name: str):
        self.name = name
        cfg = get_config("repos")
        for repo in cfg.get("repos", []):
            if repo.get("name") == self.name:
                self.config = repo
                break
        else:
            raise PluginError(f"Repo {self.name} not found in configuration.")
        self.path = get_repo_dir(self.name)
        if self.path is None or not self.path.exists():
            raise PluginError(f"No dir found for {self}, looking from {Path.cwd()}.")

    def get_plugins(self) -> List["Plugin"]:
        """Get all plugins in this repo."""
        plugins_config = self.config.get("plugin_config")
        if not plugins_config:
            raise PluginError(f"Config for {self.name} has no plugin_config key.")
        cfg = get_config(plugins_config)
        if cfg.get("plugins") is None:
            raise PluginError(f"Config {plugins_config} for {self} has no plugins key.")
        return [Plugin(repo=self, config=plugin) for plugin in cfg["plugins"]]

    def __str__(self) -> str:
        return f"Repo: {self.name} "

    def __repr__(self) -> str:
        return f"Repo(name={self.name})"


class Plugin:
    def __init__(self, repo: Repo, config: Dict[str, Any]):
        self.name = str(config.get("name"))
        if not self.name:
            raise PluginError("Plugin name is missing in the configuration.")
        self.config = config
        self.repo = repo
        self.path = self.repo.path / self.name
        if not self.path.exists():
            raise PluginError(f"Plugin directory {self.name} not found in {self.repo} expected location: {self.path}.")

    def __str__(self) -> str:
        return f"{self.repo}.{self.name})"

    def __repr__(self) -> str:
        return f"Plugin(name={self.name}, repo={self.repo})"
        pass


class DockerComposeService:
    """
    Information about a docker compose service.
    """

    def __init__(self, docker_compose_info: "DockerComposeInfo", name: str, config: Dict[str, Any]):
        self.name = name
        self.config = config
        self.docker_compose_info = docker_compose_info
        self.ports: List[PortMapping] = [PortMapping(port_mapping) for port_mapping in self.config.get("ports", [])]

    def host_name(self) -> Optional[str]:
        """Get the host name for this service."""
        if "hostname" in self.config:
            return str(self.config["hostname"])
        if "container_name" in self.config:
            return str(self.config["container_name"])
        return self.name


class DockerComposeInfo:
    """
    Information about a docker compose file.
    """

    def __init__(self, plugin: "Plugin") -> None:
        self.plugin = plugin
        self.config: Dict[str, Any]
        self.path = self.plugin.path / "docker-compose.yaml"
        if not self.path.exists():
            raise PluginError(f"Docker compose file {self.path} does not exist for plugin {self.plugin.name}.")
        self.services: Dict[str, DockerComposeService] = {}
        self.load_config()

    def load_config(self) -> None:
        """Load the docker compose configuration from the file."""
        from ruamel.yaml import YAML

        yaml = YAML(typ="safe")
        with open(self.path, "r") as file:
            self.config = yaml.load(file)
            srv = self.config.get("services", {})
            if srv is None:
                raise PluginError(
                    f"No services found in docker compose file {self.path} for plugin {self.plugin.name}."
                )
            for name, service_config in srv.items():
                self.services[name] = DockerComposeService(docker_compose_info=self, name=name, config=service_config)


def get_repos() -> List[Repo]:
    """Get all repos defined in the configuration."""
    cfg = get_config("repos")
    repos: List[Repo] = []
    for repo in cfg.get("repos", []):
        repos.append(Repo(name=repo.get("name")))
    return repos


def get_plugins() -> List[Plugin]:
    """Get all plugins defined in the configuration."""
    plugins: List[Plugin] = []
    for repo in get_repos():
        plugins.extend(repo.get_plugins())
    return plugins


def get_docker_compose_infos() -> List[DockerComposeInfo]:
    """Get all plugins defined in the configuration."""
    return [DockerComposeInfo(plugin=p) for p in get_plugins()]


def get_docker_compose_services() -> List[DockerComposeService]:
    """Get all docker compose services defined in the configuration."""
    services: List[DockerComposeService] = []
    for dc_info in get_docker_compose_infos():
        services.extend(dc_info.services.values())
    return services


if __name__ == "__main__":
    tst = get_plugins()
    for p in tst:
        dc = DockerComposeInfo(p)
        print(f"Plugin: {p.name}, Docker Compose Services: {list(dc.services.keys())}")
        for service_name, service in dc.services.items():
            print(f"  Service: {service_name}, Ports: ")
            for port in service.ports:
                print(f"      {port}")
# This code is part of the FREEDS CLI self-check plugin system.



import getpass
import secrets
import shutil
import string
from pathlib import Path
from typing import Any, Optional
from freeds.utils.root_config import RootConfig
import yaml

import freeds.utils.log as log

logger = log.setup_logging(__name__)
header_logger = log.setup_logging("FreeDS")
AUTO_YES = False

def generate_password(length: int = 8) -> str:
    alphabet = string.ascii_letters + string.digits + string.punctuation
    return "".join(secrets.choice(alphabet) for _ in range(length))


def relink(symlink: Path, target: Path) -> None:
    """Restablish a directory symlink."""
    logger.info(f'symlinking {symlink}->{target}')
    if symlink.is_symlink():
        symlink.unlink()
    symlink.symlink_to(target, target_is_directory=True)


def soft_copy(source: Path, target: Path) -> None:
    if target.exists():
        logger.info(f"{target} already exists, skipping.")
        return
    logger.info(f"copying {source} to {target}.")
    shutil.copy2(source, target)


def prompt_credential(service_desc: str, default_user: str) -> dict[str, Any]:
    """Prompt user for username and password, default username is lower case service name.
    returns a dict: {'config': {'user':'username', 'password':'user entered password'}"""

    if AUTO_YES:
        return {"config": {"user": default_user, "password": generate_password()}}

    user = input(f"Enter a username for {service_desc} [{default_user}]: ") or default_user
    password = getpass.getpass(f"Enter password for {service_desc} [auto generate]: ")
    if not password:
        password = generate_password()

    return {"config": {"user": user, "password": password}}


def prompt_press_any(description: str) -> None:
    """Print text and ask user to press any key ot continue."""
    if AUTO_YES:
        return
    input("ℹ️ " + description + "\nPress ENTER to continue")


def prompt_yesno(description: str, question: str) -> bool:
    """Provide descripiton and question, then prompt user for yes or no answer."""
    if AUTO_YES:
        return True
    prompt = f"\n❓{description}\n{question}❓ [Y/n]"
    answer = input(prompt).strip().lower()
    return answer in("y","")


def log_header(title: str, char: str, space: int = 10, vert_char: Optional[str] = None) -> None:
    """Log message in a "box" using the FreeDS logger, a bit nicer output for starting a section of actions."""
    if vert_char is None:
        vert_char = char
    width = len(title) + space * 2
    header_logger.info(char * width)
    row = vert_char + " " * (space - 1) + title + " " * (space - 1) + vert_char
    header_logger.info(row)
    header_logger.info(char * width)


def read_local_config(config_name: str) -> Optional[dict[str, Any]]:
    root_config = RootConfig()

    file_path = root_config.locals_path / (config_name + ".yaml")
    if not file_path.exists():
        return None
    with open(file_path, "r") as file:
        config: dict[str, str] = yaml.safe_load(file)
    return config


def write_local_config(config_name: str, data: dict[str, Any]) -> None:
    root_config = RootConfig()
    file_path = root_config.locals_path / (config_name + ".yaml")
    with open(file_path, "w") as file:
        yaml.dump(data, file, default_flow_style=False)


# if __name__ == '__main__':
#     log_header('This is a header', '=', 20, '-')
#     log_header('This is a header', '-')
#     log_header('This is a header', ' ')
#     logger.info('regular logging')



import logging
import os
from typing import Any, Union, cast

import requests

logger = logging.getLogger(__name__)


def get_config_url(config_name: Union[None, str] = None) -> str:
    """Get the config api server url."""
    # this must have the / at the end, otherwise the url will not work.
    base_url = os.environ.get("FREEDS_CONFIG_URL", "http://freeds-config:8005/api/configs/")
    if not base_url.endswith("/"):
        base_url += "/"
    # this can't have a slash at the end
    return f"{base_url}{config_name}" if config_name else base_url


def is_api_avaiable() -> bool:
    """Check if the config api server is available."""
    try:
        url = get_config_url()

        response = requests.head(url)
        if response.status_code != 200:
            logger.error(f"Config API server on {url} returned {response.status_code}, {response.text}")
        return response.status_code == 200
    except requests.exceptions.RequestException:
        return False


def get_full_config_response(config_name: str) -> dict[str, Any]:
    """Get a config from the config api server."""
    if config_name is None:
        raise ValueError("Config name cannot be None")

    config_url = get_config_url(config_name)

    response = requests.get(config_url)
    response.raise_for_status()
    if response.json() is None:
        raise ValueError(f"Config '{config_name}' not found. config server response: {response.text}")
    # we're basically only checking for empty files, if there is a config element we assume it to be ok.
    return cast(dict[str, Any], response.json())


def get_config_from_api(config_name: str) -> dict[str, Any]:
    """Get the a config from the api as a dict."""
    response = get_full_config_response(config_name=config_name)
    cfg = response.get("config")
    if cfg is None:
        raise ValueError(f"Config '{config_name}' does not have a 'config' key. Got json: {response}")
    # we're basically only checking for empty files, if there is a config element we assume it to be ok.
    return cast(dict[str, Any], cfg)


def get_meta(config_name: str) -> Union[None, dict[str, Any]]:
    """Get the meta data from the config api response."""
    meta = get_full_config_response(config_name=config_name).get("meta")
    return cast(dict[str, Any], meta) if meta else None



from pathlib import Path
import yaml
import os
class RootConfig:
    def freeds_file_path(self) -> Path:
        return Path.home() / ".freeds"

    def __init__(self):
        self.data:dict[str, str] = {}
        self.root_path:Path
        self.configs_path:Path
        self.locals_path:Path
        try:
            self.load()
        except (FileNotFoundError, ValueError):
            pass

    def load(self):
        p = os.environ.get('FREEDS_ROOT_PATH')
        if p:
            self.root_path = Path(p)
        else:
            with open(self.freeds_file_path(), "r") as file:
                self.data  = yaml.safe_load(file)
                cfg = self.data.get('config',{})
                self.root_path = Path(cfg.get('root_path'))
        self.configs_path = os.environ.get(
            'FREEDS_CONFIGS_PATH',
            self.root_path / "freeds-config" / "configs")
        self.configs_path = os.environ.get(
            'FREEDS_LOCALS_PATH',
            self.root_path / "local_configs")
        self.is_loaded = self.root_path is not None
        if not self.is_loaded:
            raise ValueError('could not find root config')

    def set_default(self, root_path):
        cfg = {
            "config": {
                "root_path": str(root_path)
            }
        }
        with open(self.freeds_file_path() , "w") as f:
            yaml.dump(cfg, f, default_flow_style=False)
        self.load()



"""S3 managemnent functions.
Let's see how we refactor this to transparently use config from file or api server."""

import datetime as dt
from pathlib import Path
from typing import Union

import boto3
from botocore.exceptions import ClientError

from freeds.config import get_config


def get_s3_client() -> boto3.client:
    """Get an S3 client using config from the config api"""
    cfg = get_config("s3")
    if cfg is None or cfg.get("url") is None:
        raise ValueError("s3 config not found")
    s3_client = boto3.client(
        service_name="s3",
        aws_access_key_id=cfg["access_key"],
        aws_secret_access_key=cfg["secret_key"],
        endpoint_url=cfg["url"],
    )
    return s3_client


def is_s3_service_available() -> bool:
    """Simple check if s3 works. A negative response might indicate service down or invalid credentials."""
    try:
        get_s3_client().list_buckets()
        return True
    except Exception:
        print(
            "S3 service not responding, this might be due to s3 service not started, invalid credentials or faulty/missing configuration."
        )
        return False


def bucket_exists(bucket_name: str) -> bool:
    """Check if an S3 bucket exists."""
    response = get_s3_client().list_buckets()
    for bucket in response.get("Buckets", []):
        if bucket["Name"] == bucket_name:
            return True

    return False


def file_exists(bucket_name: str, file_name: str, prefix: Union[str, None] = None) -> bool:
    """Check if a file exists on S3."""
    try:
        key = f"{prefix}/{file_name}" if prefix else file_name
        get_s3_client().head_object(Bucket=bucket_name, Key=key)
        return True
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return False
        else:
            print(f"Error checking file: {e}")
            raise


def create_bucket(bucket_name: str) -> bool:
    """Create an S3 bucket if does not exist."""

    try:
        if bucket_exists(bucket_name):
            print(f"S3 bucket {bucket_name} already exists.")
            return True
        s3_client = get_s3_client()
        s3_client.create_bucket(Bucket=bucket_name)
        print(f"Bucket {bucket_name} created.")
        return True
    except ClientError as e:
        print(f"Error creating bucket: {e}")
        return False


def delete_bucket(bucket_name: str) -> None:
    """Delete S3 bucket if it exists."""
    if not bucket_exists(bucket_name):
        return
    s3_client = get_s3_client()
    s3_client.delete_bucket(Bucket=bucket_name)


def delete_prefix(bucket: str, prefix: str) -> None:
    """Delete all files with a given prefix in a bucket."""
    s3 = get_s3_client()
    paginator = s3.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
        if "Contents" in page:
            objects = [{"Key": obj["Key"]} for obj in page["Contents"]]
            s3.delete_objects(Bucket=bucket, Delete={"Objects": objects})


def put_file(local_path: Union[str, Path], bucket: str, file_name: str, prefix: Union[str, None] = None) -> bool:
    """Upload a file to S3."""

    s3_client = get_s3_client()
    if prefix and not prefix.endswith("/"):
        prefix = prefix + "/"
    s3_key = prefix + file_name if prefix else file_name

    try:
        text = f"{local_path} to bucket '{bucket}' as '{s3_key}'"
        print(f"Intitating upload: {text}.")
        s3_client.upload_file(local_path, bucket, s3_key)
        print(f"Upload succeeded: {text}.")
        return True
    except Exception as e:
        print(f"Upload failed {text}: {e}")
        return False


def get_file(local_path: Union[str, Path], bucket: str, file_name: str, prefix: Union[str, None] = None) -> bool:
    """Download a file from S3."""
    source_object_name = f"{prefix}/{file_name}" if prefix else file_name
    text = f"{source_object_name} to {local_path}"
    print(f"Intitating download: {text}.")
    try:
        s3_client = get_s3_client()
        s3_client.download_file(bucket, source_object_name, local_path)
        print(f"Download completed: {text}.")
        return True
    except ClientError as e:
        print(f"Download failed {source_object_name} from s3: {e}")
        return False


def make_date_prefix(date: Union[dt.datetime, dt.date]) -> str:
    """Make a freeds standard prefix for the given date."""
    year = date.strftime("%Y")
    month = date.strftime("%m")
    day = date.strftime("%d")
    return f"{year}/{year}-{month}/{day}"


def as_urls(files: list[str], bucket_name: str) -> list[str]:
    return [f"s3a://{bucket_name}/{f}" for f in files]


def list_files(prefix: str, bucket_name: str) -> list[str]:
    """Expand s3 path and return all files under the given prefix, prefix should not contain any part of the filename or wildcards."""
    s3 = get_s3_client()
    paginator = s3.get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    all_files = []
    for page in pages:
        for obj in page.get("Contents", []):
            all_files.append(obj["Key"])
    return all_files


def list_files_for_dates(dates: list[Union[dt.datetime, dt.date]], root_prefix: str, bucket_name: str) -> list[str]:
    """List all files in the s3 freeds standard date paths for the given dates.
    Spark doesn't resolve wildcards so we need to list the files individually.
    Filenames are returned as haddoop compatible uri:s 's3a://bucket/prefix/filename'."""
    all_files = []
    for date in dates:
        prefix = make_date_prefix(date)
        files = list_files(f"{root_prefix}/{prefix}", bucket_name)
        all_files.extend(files)
    return all_files



"""RUnning checks defined in notebooks to get them running inside the docker network."""

import json
import os
import shutil
from importlib import resources
from pathlib import Path
from typing import List

import docker
import nbformat

from freeds.selfcheck.check_classes import (
    Check,
    CheckList,
    CheckResult,
    ExceptionCheckResult,
    MisconfiguredCheckResult,
    PluginCheckResult,
)


def get_all_notebooks() -> List[Path]:
    """
    Get a list of all notebook paths in the specified package's notebooks directory.
    """
    # Get the notebooks directory as a resources Traversable
    notebooks_dir = resources.files("freeds").joinpath("notebooks")

    # use as_file to get a Path object
    with resources.as_file(notebooks_dir) as notebooks_path:
        # we know they're not zipped so we can use the filenames
        return list(notebooks_path.rglob("*.ipynb"))


def get_result(output_nb: Path) -> CheckResult:
    """Open the output notebook, extract the result of the last cell (assumed to be JSON), and parse it to a dict."""
    nb = nbformat.read(output_nb, as_version=4)

    last_cell = nb.cells[-1]
    outputs = last_cell.get("outputs", [])
    result_dict = {}

    for output in outputs:
        # The result is usually in 'text' or 'data' fields
        try:
            if "text" in output:
                result_dict = json.loads(output["text"])
                break
            elif "data" in output and "text/plain" in output["data"]:
                result_dict = json.loads(output["data"]["text/plain"])
                break
        except json.JSONDecodeError as e:
            return ExceptionCheckResult(
                message=f"Notebook: {output_nb.name}, Invalid json in last cell:\n{output}", exception=e
            )
    if result_dict is None:
        return MisconfiguredCheckResult(
            message="Notebook: {output_nb.name}, No result found in last cell, it might not have run."
        )
    message = result_dict.get("message")
    passed = bool(result_dict.get("passed"))
    if not message or passed is None:
        return MisconfiguredCheckResult(
            message=f"Mandatory key message or passed i smissing in last cell output: {result_dict}."
        )
    plugin = result_dict.get("plugin")
    description = result_dict.get("description", "Notebook output failed to provide a description.")
    area = result_dict.get("area", "notebook")
    chk = Check(name=output_nb.name, area=area, description=description)
    result: CheckResult = (
        PluginCheckResult(passed=passed, message=message, plugin_name=plugin)
        if plugin
        else CheckResult(passed=passed, message=message)
    )
    chk.add_results(result)
    return result


def run_book(notebook_path: Path, tmp_dir: Path) -> CheckResult:
    print(f"Running notebook {notebook_path}")
    input_dir = tmp_dir / "in"
    input_dir.mkdir(parents=True, exist_ok=True)
    output_dir = tmp_dir / "out"
    output_dir.mkdir(parents=True, exist_ok=True)

    shutil.copy2(notebook_path, input_dir)

    cmd = ["papermill", f"/tmp/input/{notebook_path.name}", f"/tmp/output/{notebook_path.name}"]

    env_vars = {
        "FREEDS_CONFIG_URL": os.environ.get("FREEDS_CONFIG_URL", "http://freeds-config:8005/api/configs"),
    }

    client = docker.from_env()
    try:
        volumes = {
            str(input_dir): {"bind": "/tmp/input", "mode": "rw"},
            str(output_dir): {"bind": "/tmp/output", "mode": "rw"},
        }

        container = client.containers.run(
            image="freeds/jupyter-spark:latest",
            name=f"check-{notebook_path.name}",
            command=cmd,
            auto_remove=True,
            network="freeds-network",
            environment=env_vars,
            volumes=volumes,
            tty=False,
            detach=True,
        )

        for line in container.logs(stream=True):
            print(line.decode().strip())
        result = container.wait()
        if result.get("StatusCode", 1) != 0:
            raise RuntimeError(f"Notebook container exited with code {result['StatusCode']}")
        return get_result(output_dir / notebook_path.name)

    except docker.errors.DockerException as e:
        print(f"Docker error: {e}")
        raise
    finally:
        client.close()


def checks(tmp_dir: Path = Path("/tmp/freeds")) -> CheckList:
    check_list = CheckList("notebook")

    nb_paths = get_all_notebooks()
    nb_paths.sort()

    for nb_path in nb_paths:
        check_result = run_book(notebook_path=nb_path, tmp_dir=tmp_dir)
        if not check_result.check:
            check_result.check = Check(
                name=nb_path.name,
                description="Notebook check failed: {nb_path.name}",
                area="notebook",
                results=[check_result],
            )
        check_list.checks.append(check_result.check)

    return check_list


if __name__ == "__main__":
    run_book(
        notebook_path=Path("/Users/jens/src/FreeDS/src/freeds/notebooks/100.basics/a.hello_world.ipynb"),
        tmp_dir=Path("/tmp/freeds"),
    )



"""Configuration file access functions"""

import fcntl
from pathlib import Path
from typing import Any, Union, cast

import yaml

from freeds.config.file import freeds_root


def strip_yaml(config_name: Union[str, Path]) -> str:
    """Strip the .yaml or .yml extension from the config name."""
    if isinstance(config_name, Path):
        config_name = config_name.name

    if config_name.endswith(".yaml"):
        return config_name[:-5]
    if config_name.endswith(".yml"):
        return config_name[:-4]
    return config_name


def get_file_name(config_name: str) -> Path:
    """Get a file path for a config, searching in secrets and config folders."""
    config_name = strip_yaml(config_name)
    attempt = freeds_root() / "secrets" / (config_name + ".yaml")
    if attempt.is_file():
        return attempt
    return freeds_root() / "config" / (config_name + ".yaml")


def config_exists(config_name: str) -> bool:
    """Check if a config file exists."""
    if config_name is None:
        raise ValueError("Config name cannot be None")
    file_path = get_file_name(config_name)
    return file_path.is_file()


def read_config(config_name: str) -> dict[str, Any]:
    """Read a configuration file, returning the config data as a dict."""
    if config_name is None:
        raise ValueError("Config name cannot be None")
    file_path = get_file_name(config_name)
    if not file_path.is_file():
        raise ValueError(f"Config '{config_name}' not found in config nor secrets. Looked in {file_path}.")

    with open(file_path, "r") as file:
        config: dict[str, str] = yaml.safe_load(file)
    if not config:
        raise ValueError(f"Config '{config_name}' in {file_path} is empty or invalid.")
    if config.get("config") is None:
        raise ValueError(f"Config '{config_name}' in {file_path} does not have a 'config' key.")
    return config


def write_config_to_file(config_name: str, config: dict[str, Any]) -> None:
    """Write a configuration file, meta key is stripped if present."""
    file_path = get_file_name(config_name)
    config.pop("meta", None)
    file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(file_path, "w") as file:
        # Lock the file to prevent race conditions
        fcntl.flock(file, fcntl.LOCK_EX)
        yaml.dump(config, file, default_flow_style=False)
        fcntl.flock(file, fcntl.LOCK_UN)


def delete_config(config_name: str) -> None:
    """Delete a configuration file."""
    file_path = get_file_name(config_name)
    if file_path.exists():
        file_path.unlink()
    else:
        print(f"Configuration '{file_path}' not found.")


def list_files(path: Union[str, Path]) -> list[Path]:
    """List all files in a directory."""
    try:
        if isinstance(path, str):
            path = Path(path)
        return [f for f in path.iterdir() if f.is_file()]
    except FileNotFoundError:
        return []


def list_configs() -> list[str]:
    """List all available configurations (including the secrets)."""
    cfg_list = list_files(freeds_root() / "config")
    sec_list = list_files(freeds_root() / "secrets")
    return [strip_yaml(f.name) for f in cfg_list + sec_list if f.suffix in (".yaml", ".yml")]


def get_config_from_file(config_name: str) -> dict[str, Any]:
    """Get the a config key from a file while validating the file."""
    cfg = read_config(config_name=config_name)["config"]
    return cast(dict[str, Any], cfg)


if __name__ == "__main__":
    print(read_config("currentstack"))



import datetime as dt
import logging
import time
from typing import List, Optional, cast

from airflow_client.client.api import dag_api, dag_run_api, task_instance_api
from airflow_client.client.api_client import ApiClient
from airflow_client.client.configuration import Configuration
from airflow_client.client.model.dag_run import DAGRun
from airflow_client.client.model.task_instance import TaskInstance

from freeds.config import get_config
from freeds.selfcheck.check_classes import (
    CheckList,
    CheckResult,
    PluginCheckResult,
)

logger = logging.getLogger(__name__)


def get_dag_is_paused(dag_id: str) -> Optional[bool]:
    """Retrieve the dag enabled/disabled state from airflow."""
    with ApiClient(get_airflow_config()) as api_client:
        dag_api_instance = dag_api.DAGApi(api_client)
        dag = dag_api_instance.get_dag(dag_id)
        is_paused: bool = dag.is_paused
        return is_paused


def set_dag_is_paused(dag_id: str, is_paused: bool) -> Optional[bool]:
    """Set the dag paused/unpaused state in airflow."""
    with ApiClient(get_airflow_config()) as api_client:
        dag_api_instance = dag_api.DAGApi(api_client)
        # The API expects a dict with the key 'is_paused'
        dag_api_instance.patch_dag(dag_id, {"is_paused": is_paused})
        return is_paused == get_dag_is_paused(dag_id=dag_id)


def get_airflow_config() -> Configuration:
    """Create an airflow config objetc using the freeds config."""
    cfg = get_config("airflow")
    url = cfg.get("url")
    user = cfg.get("user")
    password = cfg.get("password")
    if not (url and user and password):
        raise ValueError("Invalid airflow config one or more values are missing (url, user or password).")
    host = f"{cfg['url']}".rstrip("/") + "/api/v1"
    cfg = Configuration(host=host, username=user, password=password)
    return cfg


def get_task_instances(dag_id: str, dag_run_id: str) -> List[TaskInstance]:
    """Get the task instances for a dag run id."""
    with ApiClient(get_airflow_config()) as api_client:
        task_instance_api_instance = task_instance_api.TaskInstanceApi(api_client)
        task_instances = task_instance_api_instance.get_task_instances(dag_id=dag_id, dag_run_id=dag_run_id)
    return cast(List[TaskInstance], task_instances.task_instances)


def get_last_dag_run(dag_id: str) -> Optional[DAGRun]:
    """Get the most recent dag run, regardless of state."""
    with ApiClient(get_airflow_config()) as api_client:
        dag_run_api_instance = dag_run_api.DAGRunApi(api_client)
        dag_runs = dag_run_api_instance.get_dag_runs(dag_id=dag_id, limit=1, order_by="-execution_date")
        if not dag_runs.dag_runs:
            logger.info("No DAG runs found.")
            return None

        latest_run = dag_runs.dag_runs[0]
        logger.info(
            f"Run ID: {latest_run.dag_run_id}, State: {latest_run.state}, Exec Date: {latest_run.execution_date}"
        )
        return latest_run


def trigger_dag_run(dag_id: str) -> DAGRun:
    """Trigger a new run of the dag."""
    with ApiClient(get_airflow_config()) as api_client:
        dag_run_api_instance = dag_run_api.DAGRunApi(api_client)
        dag_run = DAGRun(conf={"reason": "triggered via API"})
        run = dag_run_api_instance.post_dag_run(dag_id=dag_id, dag_run=dag_run)
        logger.info(f"Triggered run_id: {run.dag_run_id} at {run.execution_date}")
        return run


def check_dag_run_state(dag_id: str, dag_run_id: str) -> str:
    """Return the current dag run state for a run id."""
    with ApiClient(get_airflow_config()) as api_client:
        dag_run_api_instance = dag_run_api.DAGRunApi(api_client)
        dag_run = dag_run_api_instance.get_dag_run(dag_id=dag_id, dag_run_id=dag_run_id)
        logger.info(f"DAG Run {dag_run.dag_run_id} state: {dag_run.state}")
        return str(dag_run.state)


def trigger_and_wait_for_airflow_run(dag_id: str, timeout_seconds: int) -> CheckResult:
    """Trigger a new run of the dag and wait for it to complete or the timeout limit is reached."""
    run = trigger_dag_run(dag_id)
    # run = get_last_dag_run(dag_id)
    check_result = PluginCheckResult(passed=False, message="", plugin_name="airflow")
    timeout = dt.datetime.now() + dt.timedelta(seconds=timeout_seconds)
    while (state := check_dag_run_state(dag_id=dag_id, dag_run_id=run.dag_run_id)) in [
        "queued",
        "running",
        "scheduled",
        "up_for_retry",
    ]:
        if state != "success" and dt.datetime.now() > timeout:
            check_result.message = (
                f"Airflow Dag {dag_id} did not complete within {timeout_seconds}, it is currently in state {state}."
            )
            break
        time.sleep(3)

    check_result.passed = state == "success"
    if state == "success":
        check_result.message = f"Airflow Dag {dag_id} completed successfully."
    else:
        check_result.message = f"Airflow Dag {dag_id} completed in state {state}"

    task_instances = get_task_instances(dag_id=dag_id, dag_run_id=run.dag_run_id)
    states = "   " + "\n   ".join([f"Task: {ti.task_id} - State: {ti.state}" for ti in task_instances])

    check_result.message = check_result.message + "\n" + states
    return check_result


_dag_id = "wikipedia_pageview_pipeline"
_timeout_seconds = 5 * 60

# DAG_RUN_STATES = [
#     "queued",
#     "running",
#     "success",
#     "failed",
#     "scheduled",
#     "canceled",
#     "up_for_retry"
# ]


def check_airflow_run() -> CheckResult:
    """Run a full airflow dag using s3, spark and the full monty."""
    global _dag_id, _timeout_seconds
    is_paused = get_dag_is_paused(_dag_id)
    if is_paused:
        set_dag_is_paused(dag_id=_dag_id, is_paused=False)
    result = trigger_and_wait_for_airflow_run(dag_id=_dag_id, timeout_seconds=_timeout_seconds)
    if is_paused:
        set_dag_is_paused(dag_id=_dag_id, is_paused=True)
    return result


def checks() -> CheckList:
    """Get all checks related to web ui:s."""
    global _dag_id
    checklst = CheckList(area=__name__)
    checklst.add(
        name="airflow full run",
        description=f"run an airflow dag end-to-end, currently {_dag_id} is used",
        method=check_airflow_run,
    )
    return checklst



from typing import Callable, List, Optional, Union


class CheckResult:
    """
    Class to hold the results of a check.
    """

    def __init__(self, passed: bool, message: str, check: Optional["Check"] = None) -> None:
        self.passed = passed
        self.message = message
        self.check = check

    @property
    def symbol(self) -> str:
        return "✅" if self.passed else "❌"

    @property
    def prefix_info(self) -> str:
        """Override to insert a text before message in str and repr."""
        return ""

    @property
    def check_name(self) -> str:
        return self.check.name if self.check else "unknown"

    @property
    def check_description(self) -> str:
        return self.check.description if self.check else "unknown"

    @property
    def suffix_info(self) -> str:
        """Override to insert a text after message in str and repr."""
        return ""

    def __str__(self) -> str:
        return f"{self.symbol} - {self.check_name} - {self.prefix_info} {self.message} {self.suffix_info}"

    def __repr__(self) -> str:
        return f"{self.symbol} {self.__class__.__name__}(passed={self.passed}, check={self.check_name}, message= {self.prefix_info} {self.message} {self.suffix_info})"


class Check:
    """
    Base class for checks.
    """

    def __init__(
        self,
        name: str,
        area: str,
        description: str,
        method: Optional[Union[Callable[[], List[CheckResult]], Callable[[], CheckResult]]] = None,
        results: Optional[List[CheckResult]] = None,
    ) -> None:
        """Initiate a Check by either providing a method to execute or the result of an already executed check."""
        self.id = (area + "__" + name.lower().replace(" ", "_"),)
        self.name = name
        self.area = area
        self.description = description
        self.method = method
        self.executed = method is None
        self.results: List[CheckResult] = results if results else []

        for r in self.results:
            r.check = self

    def execute(self) -> None:
        """
        Execute the check and store the results.
        If check is marked as executed, executionis skipped.
        :return: None
        """
        if self.executed:
            print(f"Check already executed, skipping: {self.name}")
            return
        try:
            print(f"Executing check: {self.name}")
            if self.method is not None:
                results = self.method()
            self.add_results(results)

        except Exception as e:
            self.add_results(ExceptionCheckResult(f"Check '{self.name}' raised an exception.", e))
        finally:
            self.executed = True

    def add_results(self, results: Union[List[CheckResult], CheckResult]) -> None:
        if isinstance(results, list):
            self.results.extend(results)
        else:
            self.results.append(results)
        for r in self.results:
            r.check = self

    @property
    def passed(self) -> bool:
        """
        Check if all results passed.
        :return: True if all results passed, False otherwise.
        """
        if not self.executed:
            raise RuntimeError("Check has not been executed yet. Call execute() first.")
        return all(result.passed for result in self.results)

    @property
    def symbol(self) -> str:
        """
        Get the symbol representing the check result.
        :return: "✔" if passed, "✘" if failed.
        """
        if len(self.results) == 1:
            return self.results[0].symbol
        return "✅" if self.passed else "❌"

    def __str__(self) -> str:
        return f"{self.name}: {self.description}"


class PluginCheckResult(CheckResult):
    """
    The result of a check on a plugin.
    """

    def __init__(self, passed: bool, message: str, plugin_name: str) -> None:
        super().__init__(passed, message)
        self.plugin_name = plugin_name

    @property
    def prefix_info(self) -> str:
        return f"Plugin: {self.plugin_name}"


class AllGoodCheckResult(CheckResult):
    """
    Placeholder result for a check where no problems were found and thus would have no results.
    """

    def __init__(self, message: str) -> None:
        super().__init__(True, "All good, " + message)


class ExceptionCheckResult(CheckResult):
    """
    Result for a check that raised an exception.
    """

    def __init__(self, message: str, exception: Exception) -> None:
        super().__init__(False, message)
        self.exception = exception

    @property
    def symbol(self) -> str:
        return "❗️"

    def __str__(self) -> str:
        return f"{self.symbol} Exception: {self.message} - {str(self.exception)}"

    def __repr__(self) -> str:
        return f"ExceptionCheckResult(passed={self.passed}, message={self.message}, exception={str(self.exception)})"


class MisconfiguredCheckResult(PluginCheckResult):
    """
    Result for a check that has a bug in config or return values.
    """

    def __init__(self, message: str, plugin_name: Optional[str] = None) -> None:
        super().__init__(True, "Config Error: " + message, plugin_name=plugin_name or "Unknown Plugin")

    @property
    def symbol(self) -> str:
        return "❗️"


class CheckList:
    """
    Class to hold a list of checks.
    """

    def __init__(self, area: str) -> None:
        self.area = area
        self.checks: List[Check] = []

    def add(
        self, name: str, description: str, method: Union[Callable[[], List[CheckResult]], Callable[[], CheckResult]]
    ) -> None:
        """
        Add a check to the list.
        """
        self.checks.append(Check(area=self.area, name=name, description=description, method=method))

    def execute(self) -> None:
        """
        Execute all checks in the list.
        :return: List[CheckResult]
        """
        for check in self.checks:
            check.execute()

    @property
    def results(self) -> List[CheckResult]:
        """
        Get the results of all checks in the list.
        :return: List[CheckResult]
        """
        return [result for check in self.checks for result in check.results]



from .notebook import deploy_notebooks
from .stackrunner import execute_docker_compose, get_plugins
from .stackutils import (
    get_current_stack_config,
    get_current_stack_name,
    get_stack_names,
    set_current_stack,
)

__all__ = [
    "get_current_stack_config",
    "get_current_stack_name",
    "get_stack_names",
    "set_current_stack",
    "deploy_notebooks",
    "get_plugins",
    "execute_docker_compose",
]



import logging
import sys


def setup_logging(
    name: str,
    freeds_level: int = logging.INFO,
    global_level: int = logging.WARNING,
) -> logging.Logger:
    root_logger = logging.getLogger()

    if not root_logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            fmt="[%(asctime)s] %(levelname)s %(name)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    # other packages
    root_logger.setLevel(global_level)

    # our packages
    for n in [name, "freeds"]:
        logging.getLogger(n).setLevel(freeds_level)

    return logging.getLogger(name)



from .dates import date_range, parse_execution_date
from .log import setup_logging
from .root_config import RootConfig

__all__ = [
    "date_range",
    "parse_execution_date",
    "setup_logging",
    "RootConfig",
]



import datetime as dt
from pathlib import Path

from freeds.s3 import (
    bucket_exists,
    create_bucket,
    delete_bucket,
    delete_prefix,
    get_file,
    list_files,
    list_files_for_dates,
    make_date_prefix,
    put_file,
)
from freeds.selfcheck.check_classes import (
    AllGoodCheckResult,
    CheckList,
    CheckResult,
    ExceptionCheckResult,
)


def s3_check() -> CheckResult:

    bucket = "self-check-bucket"
    date = dt.date.fromisoformat("20250102")
    date_prefix = make_date_prefix(date)
    root_prefix = "s3_check"
    full_prefix = f"{root_prefix}/{date_prefix}"
    tmp_dir = Path("/tmp/freeds/selfcheck")
    source_dir: Path = tmp_dir / "source"
    source_dir.mkdir(parents=True, exist_ok=True)
    target_dir: Path = tmp_dir / "target"
    target_dir.mkdir(parents=True, exist_ok=True)
    file_name = "test.txt"

    with open(source_dir / file_name, "w") as f:
        f.write("hello s3")
    current_command = "setting up test"
    try:
        current_command = "create_bucket"
        create_bucket(bucket)

        current_command = "bucket_exists after create"
        if not bucket_exists(bucket):
            return CheckResult(
                passed=False, message=f"bucket_exists did not return true for newly created bucket {bucket}."
            )

        current_command = "put_file"

        put_file(local_path=(source_dir / file_name), bucket=bucket, prefix=full_prefix, file_name=file_name)
        expected_file_name = f"{full_prefix}/{file_name}"
        current_command = "list_files"
        files = list_files(bucket_name=bucket, prefix=full_prefix)
        if files != [expected_file_name]:
            return CheckResult(
                passed=False, message=f"Error in list_files, expected: [{expected_file_name}] got: {files}."
            )

        current_command = "list_files_for_dates"
        files = list_files_for_dates(dates=[date], root_prefix=root_prefix, bucket_name=bucket)
        if files != [expected_file_name]:
            return CheckResult(
                passed=False, message=f"Error in list_files_for_dates, expected: [{expected_file_name}] got: {files}."
            )

        current_command = "get_file"
        get_file(local_path=target_dir / file_name, bucket=bucket, prefix=full_prefix, file_name=file_name)
        if not (target_dir / file_name).is_file() or (target_dir / file_name).stat().st_size == 0:
            return CheckResult(
                passed=False,
                message=f"Could not find file: {str(target_dir / file_name)} after download or size is zero.",
            )

        current_command = "delete_prefix"
        delete_prefix(bucket=bucket, prefix=root_prefix)
        files = list_files(bucket_name=bucket, prefix=root_prefix)
        if files != []:
            return CheckResult(passed=False, message=f"Error in list_files, expected an empty list, got: {files}.")

        current_command = "delete_bucket"
        delete_bucket(bucket)

        current_command = "bucket_exists after delete"
        if bucket_exists(bucket):
            return CheckResult(passed=False, message=f"bucket_exists returns true for deleted bucket {bucket}.")

        return AllGoodCheckResult(message="Executed create+delete bucket, upload+download of file.")
    except Exception as ex:
        return ExceptionCheckResult(message=f"S3 tests failed while executing {current_command}", exception=ex)


def checks() -> CheckList:
    """Get all checks related to web ui:s."""
    checklst = CheckList(area=__name__)

    checklst.add(
        name="S3 operations check",
        description="Perform most of the s3 operations.",
        method=s3_check,
    )
    return checklst


if __name__ == "__main__":
    s3_check()



import logging
import os
from typing import Any

from freeds.config.api import get_config_from_api, is_api_avaiable
from freeds.config.file.config_classes import get_config as get_config_from_file
from freeds.config.file.config_classes import get_current_config_set
from freeds.utils import RootConfig

logger = logging.getLogger(__name__)


def get_config(config_name: str) -> dict[str, Any]:
    """Get a config, from api server if available or from file if avaiable."""
    if not config_name:
        raise ValueError("A config_name must be provided.")

    if is_api_avaiable():
        logger.debug("Using API to get config: %s", config_name)
        cfg = get_config_from_api(config_name)
        if cfg is None:
            raise FileNotFoundError(f"Config {config_name} not found in API.")
        return cfg
    else:
        logger.debug("Reading config from file: %s", config_name)
        cfg_file = get_config_from_file(config_name)
        if cfg_file is None:
            raise FileNotFoundError(f"Config {config_name} not found in files.")
        return cfg_file.get_config()


def get_env() -> dict[str, str]:
    """Get all envs as a dict.
    Root path and config url are always envs.
    Additionally any string value in the locals folder is converted to a env value, which should provide all secrets."""

    envs = {
        "FREEDS_ROOT_PATH": str(RootConfig().root_path),
        "FREEDS_CONFIGS_PATH": str(RootConfig().configs_path),
        "FREEDS_LOCALS_PATH": str(RootConfig().locals_path),
        "FREEDS_CONFIG_URL": "http://freeds-config:8005/api/configs/"
        }
    cfg_set = get_current_config_set()
    for cfg_file in (f for f in cfg_set.config_set.values() if f.is_local):
        for key, value in cfg_file.get_config().items():
            if isinstance(value, list) or isinstance(value, dict):
                continue
            env_name = f"FREEDS_{cfg_file.config_name.upper()}_{key.upper()}"
            envs[env_name] = str(value)
    return envs


def set_env() -> None:
    """set all env values"""
    for key, value in get_env().items():
        os.environ[key] = value


get_env()



import typer

from freeds.cli.helpers.stackutils import get_current_stack_name, set_current_stack
from freeds.config import get_config

cfg_app = typer.Typer(help="Manage freeds stacks.")


@cfg_app.command()  # type: ignore
def ls() -> None:
    """List all stacks."""

    cfg = get_config("stacks")
    current_stack = get_current_stack_name()
    if current_stack is None:
        typer.echo("No current stack set,use 'freeds setstack <name>' to set one.")
    for stack in cfg.keys():
        if stack == current_stack:
            typer.echo(f"** stack: {stack} ** (current)")
        else:
            typer.echo(f"stack: {stack}")
        for service in cfg[stack].get("plugins", []):
            typer.echo(f"  - {service}")


@cfg_app.command()  # type: ignore
def set(
    stack: str = typer.Argument(..., help="Stack name to set as current"),
) -> None:
    """Set freeds to use the provided stack."""
    set_current_stack(stack_name=stack)



"""Assume all ports 8000-8999 are http servers, if the plugin does not respond by http, the port should be changed on the plugin"""

from typing import List

import requests

from freeds.selfcheck.check_classes import (
    AllGoodCheckResult,
    CheckList,
    CheckResult,
    PluginCheckResult,
)
from freeds.selfcheck.plugin_classes import get_docker_compose_services


def check_web_uis_localhost() -> List[CheckResult]:
    """
    Check if the web UI is running and accessible.
    :return: None
    """
    result: List[CheckResult] = []
    for service in get_docker_compose_services():
        for port_mapping in service.ports:
            if 8000 <= port_mapping.host_port <= 8999:
                try:
                    url = f"http://127.0.0.1:{port_mapping.host_port}"
                    response = requests.head(url)
                    result.append(
                        PluginCheckResult(
                            passed=True,
                            message=f"Http response for url {url} is {response.status_code}. {service.docker_compose_info.plugin}, service {service.name}.",
                            plugin_name=service.docker_compose_info.plugin.name,
                        )
                    )
                except requests.exceptions.RequestException:
                    result.append(
                        PluginCheckResult(
                            passed=False,
                            message=f"Failed to connect to url {url}. {service.docker_compose_info.plugin} service: {service.name}.",
                            plugin_name=service.docker_compose_info.plugin.name,
                        )
                    )
    return result


def check_port_mapped_to_localhost() -> List[CheckResult]:
    """
    Check if the ports are mapped to localhost.
    :return: List of CheckResult objects.
    """
    result: List[CheckResult] = []
    for service in get_docker_compose_services():
        for port_mapping in service.ports:
            if port_mapping.host_ip != "127.0.0.1":
                result.append(
                    PluginCheckResult(
                        passed=False,
                        message=f"Service {service.name} is not mapped to localhost, host IP is {port_mapping.host_ip}.",
                        plugin_name=service.docker_compose_info.plugin.name,
                    )
                )
    if len(result) == 0:
        result.append(
            AllGoodCheckResult(
                message="All ports are mapped to 127.0.0.1, i.e. no conteiner ports accessible from outside.",
            )
        )
    return result


def checks() -> CheckList:
    """Get all checks related to web ui:s."""
    checklst = CheckList(area=__name__)

    checklst.add(
        name="HTTP server health",
        description="Check if web UIs and other http servers are running and accessible, accepting any http response as valid.",
        method=check_web_uis_localhost,
    )
    checklst.add(
        name="Ports mapped to localhost",
        description="Check that no ports mapped fgrom conatiners are accessible from outsid the host machine, i.e. all ports are mapped to 127.0.0.1:nnnn",
        method=check_port_mapped_to_localhost,
    )
    return checklst



import typer

from freeds.cli.helpers import execute_docker_compose, get_plugins

app = typer.Typer()


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})  # type: ignore[misc]
def dc(
    ctx: typer.Context,
    single: str = typer.Option(None, "-s", "--single", help="Folder to run docker compose in"),
    # args: list[str] = typer.Argument(None, help="Arguments to pass to docker compose")
) -> int:
    """
    Call docker compose with the supplied parameters for all freeds plugins in the current stack.
    """
    args = ctx.args
    if not args:
        print("Error: docker compose command must be given")
        return 1

    plugins: list[str] = []
    if single is None:
        plugins = get_plugins()
        print(f"Found plugins: {plugins}")
    else:
        plugins = [single]

    execute_docker_compose(params=args, plugins=plugins)
    return 0



from .s3 import (
    as_urls,
    bucket_exists,
    create_bucket,
    delete_bucket,
    delete_prefix,
    file_exists,
    get_file,
    get_s3_client,
    is_s3_service_available,
    list_files,
    list_files_for_dates,
    make_date_prefix,
    put_file,
)

__all__ = [
    "bucket_exists",
    "file_exists",
    "create_bucket",
    "delete_bucket",
    "delete_prefix",
    "get_file",
    "get_s3_client",
    "is_s3_service_available",
    "list_files",
    "list_files_for_dates",
    "make_date_prefix",
    "put_file",
    "as_urls",
]



from typing import List

import typer

from freeds.selfcheck import (
    airflow_checks,
    directory_checks,
    docker_checks,
    network_checks,
    notebook_checks,
    s3_checks,
)
from freeds.selfcheck.check_classes import (
    CheckList,
    CheckResult,
    ExceptionCheckResult,
)


def selfcheck(
    no_nb: bool = typer.Option(False, "--no-nb", help="Skip the notebook based checks(spark)."),
    no_airflow: bool = typer.Option(False, "--no-airflow", help="Skip the airflow check."),
) -> int:
    """
    Perform all self checks.
    """

    checklists: List[CheckList] = [
        docker_checks.checks(),
        directory_checks.checks(),
        network_checks.checks(),
        s3_checks.checks(),
    ]
    if not no_nb:
        checklists.append(notebook_checks.checks())
    if not no_airflow:
        checklists.append(airflow_checks.checks())

    results: List[CheckResult] = []
    for checklist in checklists:
        try:
            checklist.execute()
            results.extend(checklist.results)
        except Exception as e:
            results.append(ExceptionCheckResult(message="A checklist execution raised an exception.", exception=e))

    for result in results:
        typer.echo(result)
    return 0


if __name__ == "__main__":
    selfcheck(no_airflow=True, no_nb=True)



import typer

from freeds.cli.commands import dc, env, nb, selfcheck, stack
from freeds.config import set_env

set_env()


app = typer.Typer()

app.add_typer(dc.app)  # <-- Add this line)
app.command()(env.env)
app.command()(selfcheck.selfcheck)
app.add_typer(nb.nb_app, name="nb")
app.add_typer(stack.cfg_app, name="stack")


if __name__ == "__main__":
    app()



import os
from pathlib import Path
from typing import Any, Optional

import yaml

import freeds.utils.log as log
from freeds.utils import RootConfig

logger = log.setup_logging(__name__)


class ConfigFile:
    """Class to facade single config file"""

    def __init__(self, file_path: Path, config_set: "ConfigSet", source:str) -> None:
        self.source = source
        self.config_set = config_set
        self.source_file_path = file_path
        self.data: dict[str,Any] = None
        self.load()

    @property
    def config_name(self):
        return self.source_file_path.stem

    @property
    def is_local(self)->bool:
        return self.source=='locals'

    def load(self) -> None:
        if not self.source_file_path.exists():
            raise FileNotFoundError(f"Config file {self.source_file_path} does not exists.")
        with open(self.source_file_path, "r") as file:
            data: dict[str, Any] = yaml.safe_load(file)
            self.data = data
        self.validate()
        if self.data is None:
            raise ValueError(f"Config file malformed or empty (data is None) {self.source_file_path}")

    def validate(self, raise_for_error: bool = True) -> bool:
        """Check that format is valid, returns true if data is None."""
        message = None
        if self.data is None:
            return True

        if message is None and not self.data.get("config"):
            message = "The config has no 'config' root key."

        if raise_for_error and message:
            raise (ValueError(message))
        return message is None

    def get_config(self) -> dict[str, Any]:
        """Get the content of the "config" element in the data"""
        if not self.data:
            self.load()
        data: dict[str, Any] = self.data["config"]  # type: ignore[index]
        return data

class ConfigSet:
    """Class for scanning a single set of config files.
    locals override configs
    """

    def __init__(self, configs_path: Path, locals_path: Path) -> None:
        self.configs_path = configs_path
        self.locals_path = locals_path

        configs = self.list_files(path=configs_path, source='configs')
        locals =  self.list_files(path=locals_path, source='locals')
        self.config_set = configs | locals

    def list_files(self, path:Path, source:str) -> dict[str, ConfigFile]:
        result: dict[str, ConfigFile]= {}
        for f in path.iterdir():
            if not (f.suffix in {".yaml", ".yml"} and f.is_file()):
                continue
            cfg = ConfigFile(file_path=f, config_set=self, source=source)
            result[cfg.config_name] = cfg
        return result



def freeds_config_set() -> ConfigSet:
    """Get the freeds config set (from config folder in the root freeds folder)."""
    cfg = RootConfig()
    cfg_set = ConfigSet(configs_path=cfg.configs_path, locals_path=cfg.locals_path)
    return cfg_set


def get_current_config_set() -> ConfigSet:
    """get all configured configs, which for now is only freeds"""
    return freeds_config_set()


def get_config(config_name: str) -> Optional[ConfigFile]:
    """Get config object for the config_name"""
    cfg_set = get_current_config_set().config_set
    cfg = cfg_set.get(config_name)
    if cfg:
        cfg.load()
    return cfg



import datetime as dt
from typing import Union


def parse_execution_date(execution_date: Union[str, dt.date, dt.datetime]) -> dt.datetime:
    """Parse an airflow execution (iso format) date to a datetime object, accepts date and datetime as well, always returning a datetime."""
    if isinstance(execution_date, dt.datetime):
        return execution_date
    if isinstance(execution_date, dt.date):
        # Convert date to datetime at midnight
        return dt.datetime.combine(execution_date, dt.time())

    return dt.datetime.fromisoformat(execution_date)


def date_range(execution_date: Union[str, dt.datetime, dt.date], length: int) -> list[dt.datetime]:
    """Return a series of <length> datetimes, ending with execution_date."""
    start_date = parse_execution_date(execution_date)
    return [start_date - dt.timedelta(days=i) for i in reversed(range(length))]



import os
import subprocess
from pathlib import Path
from typing import List, Optional, cast

import freeds.utils.log as log
from freeds.cli.helpers.stackutils import (
    get_current_stack_config,
    get_current_stack_name,
)
from freeds.config import set_env
from freeds.utils import RootConfig

logger = log.setup_logging(__name__)


def get_plugins() -> Optional[List[str]]:
    set_env()
    current_stack = get_current_stack_name()

    if current_stack is None:
        print("Error: No current stack set. Use `freeds setstack <stackname>` to set a stack.")
        return None

    current_stack_cfg = get_current_stack_config()
    if current_stack_cfg is None:
        print(f"Error: No configuration found for current stack '{current_stack}'.")
        return None

    plugins = current_stack_cfg.get("plugins")
    if not plugins:
        print(f"Error: malformed config, 'plugins' key is missing on current stack '{current_stack}'.")
        return None

    return cast(List[str], plugins)


def execute_docker_compose(params: List[str], plugins: List[str]) -> None:
    set_env()
    start_dir = Path.cwd()

    run_in_current_dir = plugins == ["."]

    plugin_root = RootConfig().root_path / "the-free-data-stack"
    command = params[0]
    if command in ["down", "stop"]:
        plugins = list(reversed(plugins))
    if command in ["up", "start"] and "-d" not in params:
        params.append("-d")

    dc = ["docker", "compose", *params]

    # Execute the command for each plugin
    print(f"Running '{' '.join(dc)}' for plugins: {plugins}")
    for plugin in plugins:
        if not run_in_current_dir:
            plugin_dir = plugin_root / plugin
            if not plugin_dir.exists():
                print(f"Warning: Plugin directory '{plugin_dir}' does not exist. Skipping.")
                continue

            os.chdir(plugin_dir)

        try:
            print(f"Executing in: {Path.cwd()}")
            subprocess.run(dc, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error: Failed to execute 'docker compose {command}' for plugin '{plugin}':{e}.")
        finally:
            os.chdir(start_dir)



from typing import Any

import docker

from freeds.config import get_config
from freeds.selfcheck.check_classes import (
    AllGoodCheckResult,
    CheckList,
    CheckResult,
    MisconfiguredCheckResult,
    PluginCheckResult,
)


def find_key(data: Any, target_key: str) -> Any:
    """
    Recursively traverse a nested dict/list and yield all values for target_key.
    """
    if isinstance(data, dict):
        for key, value in data.items():
            if key == target_key:
                yield value
            if isinstance(value, (dict, list)):
                yield from find_key(value, target_key)
    elif isinstance(data, list):
        for item in data:
            yield from find_key(item, target_key)


def get_merged_list(data: dict[str, Any], key: str) -> list[str]:

    lists = find_key(data, key)
    values = []
    for value in lists:
        if isinstance(value, list):
            values.extend(value)
        else:
            print(f"Warning: Expected only lists for key '{key}', but got {type(value).__name__}.")
    return values


def check_duplicate_ports() -> list[CheckResult]:
    """
    Check if there are duplicate ports in the plugins configuration.
    :return: List of PluginCheckResult indicating if there are duplicate ports.
    """
    cfg = get_config("plugins")
    results: list[CheckResult] = []
    port_index: dict[str, Any] = {}
    for plugin in cfg.get("plugins", []):
        ports = plugin.get("ports", [])
        for port in ports:
            port_number = port.get("number")
            if not port_number:
                results.append(
                    MisconfiguredCheckResult(
                        message=f"Port number is missing in config: '{port}'.", plugin_name=plugin.get("name")
                    )
                )
                continue
            pl = port_index.get(port_number)
            if not pl:
                port_index[port_number] = [plugin]
            else:
                pl.append(plugin)

    for key, value in port_index.items():
        if len(value) == 1:
            continue
        pluginstr = ", ".join([p.get("name") for p in value])
        for plugin in value:
            results.append(
                PluginCheckResult(False, f"Duplicate port found in {pluginstr}: {key}", plugin_name=plugin.get("name"))
            )
    if len(results) == 0:
        results.append(AllGoodCheckResult("No duplicate ports found."))

    return results


def check_duplicate_container_names() -> list[CheckResult]:
    """
    Check if there are duplicate containers in the plugins configuration.
    :return: List of PluginCheckResult indicating if there are duplicate containers.
    """
    cfg = get_config("plugins")
    results: list[CheckResult] = []
    container_index: dict[str, Any] = {}
    for plugin in cfg.get("plugins", []):
        containers = plugin.get("containers", [])
        for container in containers:
            pl = container_index.get(container)
            if not pl:
                container_index[container] = [plugin]
            else:
                pl.append(plugin)

    for key, value in container_index.items():
        if len(value) == 1:
            continue
        pluginstr = ", ".join([p.get("name") for p in value])
        for plugin in value:
            results.append(
                PluginCheckResult(
                    False, f"Duplicate container found in {pluginstr}: {key}", plugin_name=plugin.get("name")
                )
            )
    if len(results) == 0:
        results.append(AllGoodCheckResult("No duplicate containers found."))

    return results


def check_missing_containers() -> list[CheckResult]:
    """
    Check if all containers in the list are running.
    :param container_names: List of container names or IDs.
    :return: True if all are running, False otherwise.
    """
    cfg = get_config("plugins")
    running_containers = [
        container.name for container in docker.from_env().containers.list() if container.status == "running"
    ]
    results: list[CheckResult] = []
    for plugin in cfg.get("plugins", []):
        containers = plugin.get("containers", [])
        if not containers:
            results.append(
                PluginCheckResult(
                    False,
                    f"No containers specified in plugins.yaml for plugin'{plugin.get('name')}'.",
                    plugin_name=plugin.get("name"),
                )
            )
            continue
        for container in plugin.get("containers", []):
            if container in running_containers:
                results.append(
                    PluginCheckResult(True, f"Container '{container}' is running.", plugin_name=plugin.get("name"))
                )
            else:
                results.append(
                    PluginCheckResult(
                        False, f"Expected container '{container}' is not running.", plugin_name=plugin.get("name")
                    )
                )
    return results


def checks() -> CheckList:
    """Run all checks related to docker."""
    checklst = CheckList(__name__)
    checklst.add(
        name="Expected Docker Containers",
        description="Check if all expected Docker containers are running.",
        method=check_missing_containers,
    )

    checklst.add(
        name="Duplicate Ports",
        description="Check that no plugins claim the same ports.",
        method=check_duplicate_ports,
    )

    checklst.add(
        name="Duplicate Container Names",
        description="Check that no plugins claim the same container name.",
        method=check_duplicate_container_names,
    )
    return checklst



import typer

from freeds.cli.helpers import deploy_notebooks
from freeds.config import get_config
from freeds.s3 import delete_prefix, list_files

nb_app = typer.Typer(help="Manage notebooks on S3.")


@nb_app.command()  # type: ignore
def deploy(
    repo: str = typer.Option("all", "--repo", help="Deploy a single repo only."),
    normalize: bool = typer.Option(False, "--normalize", help="Normalize source file before deploying."),
) -> None:
    """Deploy notbooks to S3 from repo(s), timestamping and tagging them with git revision.
    Optionally you can use this command to upgrade the notebook formats to 4.5, silencing the id warning."""
    if normalize:
        print("Normalizing source files before deployment...")
    deploy_notebooks(repo=repo, normalize_source=normalize)


@nb_app.command()  # type: ignore
def cfg() -> None:
    """Show repo config."""
    cfg = get_config("nbdeploy")
    for repo in cfg.get("repos", []):
        print(f"Repo: {repo['name']}")
        print("Directories:")
        for folder in repo.get("directories", []):
            print(f"  - {folder}")


@nb_app.command()  # type: ignore
def ls(prefix: str = typer.Argument(None, help="List files under this prefix")) -> None:
    """List all notebooks on S3."""
    cfg = get_config("nbdeploy")
    bucket = cfg.get("bucket", "notebooks")
    if prefix:
        print(f"Files under prefix '{prefix}' in bucket {bucket}:")
        books = list_files(bucket_name=bucket, prefix=prefix)
        for book in books:
            print(f"  - {book}")
        return

    for repo in cfg.get("repos", []):
        books = list_files(bucket_name="notebooks", prefix=repo["name"])
        print(f"Repo {repo['name']} has {len(books)} files in bucket {bucket}:")
        for book in books:
            print(f"  - {book}")


@nb_app.command()  # type: ignore
def delprefix(prefix: str = typer.Argument(..., help="Delete all files under this prefix")) -> None:
    """Delete all files under an S3 prefix."""
    cfg = get_config("nbdeploy")
    bucket = cfg.get("bucket", "notebooks")
    books = list_files(bucket_name="notebooks", prefix=prefix)
    for book in books:
        print(f"  - {book}")
    if not typer.confirm(f"{len(books)} files under prefix {prefix} in bucket {bucket} will be deleted. Continue?"):
        print("Deletion cancelled.")
        return
    delete_prefix(bucket=bucket, prefix=prefix)


# deploy("all")
# delprefix("pipe-dreams")



import pyspark
from delta import configure_spark_with_delta_pip
from pyspark.sql import SparkSession

from freeds.config import get_config


def get_spark_session(app_name: str, use_local: bool = False) -> SparkSession:
    """Get spark client for s3."""

    s3_cfg = get_config("s3")

    conf = (
        pyspark.conf.SparkConf()
        .setAppName(app_name)
        # s3 secrets
        .set("spark.hadoop.fs.s3a.access.key", s3_cfg["access_key"])
        .set("spark.hadoop.fs.s3a.secret.key", s3_cfg["secret_key"])
        .set("spark.task.maxFailures", "1")
    )
    if use_local:
        conf = conf.setMaster("local[*]")

    builder = pyspark.sql.SparkSession.builder.config(conf=conf)
    spark_session = configure_spark_with_delta_pip(builder).getOrCreate()

    return spark_session


def show_cfg(spark_session: SparkSession) -> None:
    """Print the entire spark config."""
    cfg = spark_session.sparkContext.getConf().getAll()
    for key, value in cfg:
        if key in (
            "spark.submit.pyFiles",
            "spark.driver.extraJavaOptions",
            "park.app.initial.jar.urls",
            "spark.files",
            "spark.repl.local.jars",
            "spark.app.initial.file.urls" "spark.executor.extraJavaOption",
            "spark.app.initial.jar.urls" "spark.app.initial.file.urls",
        ):
            print(key)
            for csv in value.split(","):
                print("    " + str(csv))
        else:
            print(f"{key} = {value}")


def show_spark_info(sc: SparkSession) -> None:
    """Print some spark info."""
    cfg: pyspark.SparkConf = sc.sparkContext.getConf()
    print(f'==== spark app: {cfg.get("spark.app.name")} ====')
    print(f'Spark master: {cfg.get("spark.master")}')
    print(f'Delta lake location: {cfg.get("spark.sql.warehouse.dir")}')
    print(f'S3 endpoint: {cfg.get("spark.hadoop.fs.s3a.endpoint")}')


def show_dbs(sc: SparkSession) -> None:
    """Print all databases and tables."""
    dbs = sc.catalog.listDatabases()
    print("Databases and tables:")
    for db in dbs:
        print(db.name)
        tables = sc.catalog.listTables(db.name)
        for tbl in tables:
            print(f"    {tbl.name}")



#!/usr/bin/env python3
from typing import Any, Optional, Union, cast

from freeds.config import get_config
from freeds.setup.utils import write_local_config
def get_current_stack_name() -> Union[None, str]:
    """Get the current stack name from the currentstack.yaml file."""
    cur_stack = get_config("currentstack")
    if not cur_stack:
        return None
    stack = cur_stack.get("current_stack")
    return None if stack is None else str(stack)


def get_stack_config(stack_name: str) -> Optional[dict[str, Any]]:
    if stack_name is None:
        return None
    cfg = get_config("stacks")
    return cast(dict[str, Any], cfg.get(stack_name))


def get_current_stack_config() -> Optional[dict[str, Any]]:
    """Get the current stack config from the currentstack.yaml file."""
    cur_stack = get_current_stack_name()
    if not cur_stack:
        return None
    return get_stack_config(cur_stack)


def get_plugins(current_only: bool = True) -> list[dict[str, Any]]:
    """Get a list the plugin info for current stack or all plugins."""
    plugins = get_config("plugins").get("plugins", [])
    if not current_only:
        return cast(list[dict[str, Any]], plugins)

    current_cfg = get_current_stack_config()
    if not current_cfg:
        raise ValueError("No current stack set or no stack config defined.")
    current_plugins = current_cfg.get("plugins", [])

    return list(p for p in plugins if p["name"] in current_plugins)


def get_stack_names() -> list[str]:
    """Get a list of stacknames"""
    stacks = get_config("stacks")
    return list(stacks.keys())


def set_current_stack(stack_name: str) -> None:
    """Set freeds to use the provided stack."""
    stack_found = False
    for name in get_stack_names():
        if stack_name == name:
            stack_found = True
            break
    if not stack_found:
        print(f"Error: Stack '{stack_name}' not found, use `freeds ls` to see available stacks.")

    # Lock the file to prevent race conditions
    config = {
        "annotation": "the current stack for freeds cli, use setstack to change it, editing here is fine too",
        "config": {"current_stack": stack_name},
    }

    write_local_config(config_name="currentstack", data=config)
    print(f"Current stack set to '{stack_name}'.")



from pathlib import Path
from typing import Any

import freeds.setup.utils as utils
import freeds.utils.log as log

logger = log.setup_logging(__name__)


def merge_config(config_name: str, new_cfg: dict[str, Any]) -> None:
    """Load old config if exists, update with new values and save it, otherwise save the new_cfg as is."""
    old_cfg = utils.read_local_config(config_name=config_name)
    if old_cfg:
        old_cfg.update(new_cfg)
        new_cfg = old_cfg
    utils.write_local_config(config_name=config_name, data=new_cfg)
    logger.info(f"✅ Updated local config {config_name}.yaml with new credentials.")


def setup_credentials() -> bool:
    """Prompt user for credentials and save them in locals."""
    utils.log_header(title="Create credentials for airflow, S3 etc", char="-")
    utils.prompt_press_any(
        "You'll be prompted for credentials, just press enter to use default user name and an autogenerated 8 char password.\n"
        "The credentials will be readable in the config files in the 'config/locals' folder.\n"
        "Nothing in FreeDS should be exposed outside your computer, so security is not a huge concern. On the other hand, you never know. I'm human, bugs happen...\n"
    )

    airflow_cfg = utils.prompt_credential(service_desc="Airflow", default_user="freeds")
    merge_config(config_name="airflow", new_cfg=airflow_cfg)

    s3_cfg = utils.prompt_credential(service_desc="minio S3", default_user="freeds")
    merge_config(config_name="s3", new_cfg=s3_cfg)
    merge_config(config_name="minio", new_cfg=s3_cfg)

    postgres_cfg = utils.prompt_credential(service_desc="PostgreSQL", default_user="freeds")
    merge_config(config_name="postgres", new_cfg=postgres_cfg)
    utils.log_header(title="🟢 Credentials setup completed successfully 🌟", char=" ")
    return True


# if __name__ == '__main__':

#     setup_credentials()



import docker

import freeds.setup.utils as utils
import freeds.utils.log as log

logger = log.setup_logging(__name__)


def is_docker_running() -> bool:
    try:
        client = docker.from_env()
        client.ping()  # Will raise an exception if Docker is not running
        return True
    except Exception as s:
        print(s)
        return False


def prompt_try_again() -> bool:
    return utils.prompt_yesno(description="The Docker daemon is not running.", question="Fix it and we try again?")


def setup_docker() -> bool:
    utils.log_header("Setting up FreeDS docker", "-")
    while not is_docker_running():
        if not prompt_try_again():
            return False
    network_name = "freeds-network"
    client = docker.from_env()

    networks = client.networks.list(names=[network_name])
    if networks:
        logger.info(f"✅ Network '{network_name}' already exists.")
    else:
        logger.info(f"Creating the docker network '{network_name}'")
        client.networks.create(network_name, driver="bridge")
        logger.info(f"✅ Network '{network_name}' created.")
    utils.log_header(title="🟢 Docker setup completed successfully 🌟", char=" ")
    return True


if __name__ == "__main__":
    setup_docker()



import freeds.setup as setup
import freeds.setup.utils as utils
import freeds.utils.log as log

logger = log.setup_logging(__name__)


def main() -> bool:
    utils.log_header("Setting up The Free Data Stack (FreeDS)", "=")
    # Directory setup
    if not setup.setup_root_dir():
        return False
    if not utils.prompt_yesno(description="FreeDS directory setup is completed", question="continue to docker setup"):
        return False

    # Docker setup
    if not setup.setup_docker():
        return False
    if not utils.prompt_yesno(description="FreeDS docker setup is completed", question="continue to credentials setup"):
        return False

    # Credentials setup
    if not setup.setup_credentials():
        return False
    if not utils.prompt_yesno(
        description=(
            "Credential setup is completed, next step is to create and start the FreeDS plugins in docker.\n"
            "This is the point of no return in regards to some of the config. The process so far can be re-run and new credentials entered and so on.\n"
            "Next step will create users from the credentials and they cannot be updated by simply modifiying the config files."
        ),
        question="Move on to initilise the FreeDS plugins",
    ):
        return False

    return True



from .config_api import (
    get_config_from_api,
    get_config_url,
    get_full_config_response,
    get_meta,
    is_api_avaiable,
)

__all__ = [
    "get_config_from_api",
    "get_meta",
    "is_api_avaiable",
    "get_config_url",
    "get_full_config_response",
]



import datetime as dt
import os
import shutil
import sys
from pathlib import Path
from typing import Any, Optional, cast

import git
import nbformat

from freeds.config import get_config
from freeds.s3 import put_file


def find_dir(dir_name: str) -> Path:
    """Find the specified directory in the current working directory or its parent directories."""
    looked_in = []
    p = Path(dir_name)
    # find repo and notebooks directories
    for i in range(4):
        looked_in.append(p.absolute())
        if p.exists():
            return p.resolve()
        p = Path("..") / p
    else:
        raise FileNotFoundError(f"Error: directory '{dir_name}' not found, looked in:\n{looked_in}")


def get_repo_config(repo: str) -> dict[str, Any]:
    cfg = get_config("nbdeploy")
    for r in cfg.get("repos", []):
        if r.get("name") == repo:
            return cast(dict[str, Any], r)
    return {}


def get_git_info() -> dict[str, str]:
    """Get Git information using GitPython."""
    try:
        repo = git.Repo(".")
        head = repo.head.commit
        url = repo.remotes.origin.url
        if url is None:
            url = "No remote URL found"
        else:
            url = f"{url.rstrip('.git')}/commit/{head.hexsha}"
        return {
            "repo": repo.working_tree_dir,
            "branch": repo.active_branch.name,
            "revision": head.hexsha[:7],  # Short hash
            "commit_date": head.committed_datetime.isoformat(),
            "author": f"{head.author.name}",
            "deployed": dt.datetime.now(dt.timezone.utc).isoformat(),
            "url": url,
        }
    except git.InvalidGitRepositoryError:
        print("Error: Not a git repository")
        sys.exit(1)
    except Exception as e:
        print(f"Error getting git info: {e}")
        sys.exit(1)


def format_md(git_info: dict[str, str], input_path: str) -> str:
    """
    Format the Git information into a markdown string.
    """
    # Create the revision markdown content
    url = git_info["url"]
    return (
        f"# Notebook: {os.path.basename(input_path)}\n\n"
        f"> **Git Revision**: `{git_info['revision']}` | **Branch**: `{git_info['branch']}`\n\n"
        f"> **Commit Date**: {git_info['commit_date']} | **Author**: {git_info['author']}\n\n"
        f"> **Deployed**: {git_info['deployed']}\n\n"
        f"> [{url}]({url})"
    )


def find_cell_by_tag(notebook: nbformat.NotebookNode, tag: str) -> nbformat.NotebookNode:
    """
    Find a cell in the notebook with the specified tag.
    Returns the cell if found, otherwise None.
    """
    for cell in notebook.cells:
        if "metadata" in cell and "tags" in cell.metadata and tag in cell.metadata.tags:
            return cell
    return None


class NotebookFile:
    def __init__(self, repo: str, repo_path: str, notebook_dir: str, dir: str, file: str) -> None:
        self.repo = repo
        self.notebook_dir = notebook_dir
        self.repo_path = repo_path
        self.dir = None if dir == "." else dir
        self.file = file

    def s3_prefix(self) -> str:
        """
        Return the S3 prefix for this notebook file.
        """
        if self.dir:
            return f"{self.repo}/{self.dir}"
        return f"{self.repo}/{self.notebook_dir}"

    def local_prefix(self) -> str:
        """
        Return the directory path for this notebook file relative (and including) the repository.
        """
        if self.dir:
            return os.path.join(self.repo, self.dir)
        return os.path.join(self.repo, self.notebook_dir)

    def temp_file_path(self, temp_dir: str) -> str:
        """
        Return the local path for this notebook file in the temporary directory.
        """
        return os.path.join(temp_dir, self.local_prefix(), self.file)

    def local_file_path(self) -> str:
        """
        Return the local path for this notebook file.
        """
        if self.dir:
            return os.path.join(self.repo_path, self.dir, self.file)
        else:
            return os.path.join(self.repo_path, self.notebook_dir, self.file)

    def s3_file_path(self) -> str:
        """
        Return the s3 path for this notebook file.
        """
        return os.path.join(self.s3_prefix(), self.file)


def list_files(repo: str, notebook_dir: str) -> list[NotebookFile]:
    """
    List all notebook files in the specified directory of the repository.
    Returns a list of NotebookFile objects.
    """
    repo_dir = find_dir(repo)
    if repo_dir is None:
        print(f"Error: repository '{repo}' not found")
        return []

    notebook_path = os.path.join(repo_dir, notebook_dir)
    if not os.path.exists(notebook_path):
        print(f"Notebook directory '{notebook_dir}' not found in '{repo_dir}'")
        return []

    notebook_files = []
    for dirpath, _, files in os.walk(notebook_path):
        rel_dir = os.path.relpath(dirpath, start=repo_dir)
        for file in files:
            if file.endswith(".ipynb"):
                notebook_files.append(
                    NotebookFile(repo=repo, repo_path=str(repo_dir), notebook_dir=notebook_dir, dir=rel_dir, file=file)
                )
    return notebook_files


def stamp_notebook(input_path: str, output_path: str) -> Optional[nbformat.NotebookNode]:
    """
    Add Git revision information to notebook metadata and2
    insert/update a markdown cell at the top with revision info.
    Uses cell tags to identify the Git info cell.
    """
    if Path(input_path).resolve() == Path(output_path).resolve():
        raise ValueError("Input and output paths must be different to avoid overwriting the original notebook.")
    try:
        git_info = get_git_info()
        # Load the notebook
        with open(input_path, "r", encoding="utf-8") as f:
            notebook = nbformat.read(f, as_version=4)

        # Look for a cell with the 'gitinfo' tag
        revision_cell = find_cell_by_tag(notebook, "gitinfo")
        if revision_cell is None:
            # If no existing cell found, create a new one
            revision_cell = nbformat.v4.new_markdown_cell()
            notebook.cells.insert(0, revision_cell)

        # Set the gitinfo on the cell metadata and content
        if "metadata" not in revision_cell:
            revision_cell["metadata"] = {}
        if "tags" not in revision_cell["metadata"]:
            revision_cell["metadata"]["tags"] = []
        revision_cell["metadata"]["tags"].append("gitinfo")
        revision_cell["source"] = format_md(git_info, input_path)

        print(f"Writing a copy of the notebook to: {output_path}.")
        containing_dir = os.path.dirname(output_path)
        os.makedirs(containing_dir, exist_ok=True)
        with open(output_path, "w", encoding="utf-8") as f:
            nbformat.write(notebook, f)
        print(f"Notebook {input_path} stamped with Git info and saved to {output_path}.")
        return notebook
    except Exception as e:
        print(f"Error processing notebook {input_path}: {e}")
        return None


def normalize(files: list[NotebookFile]) -> None:
    """
    Normalize the source files to verrsion 4.5.
    """
    for file in files:
        print(f"Normalizing notebook: {file.local_file_path()}.")
        with open(file.local_file_path(), "r", encoding="utf-8") as f:
            nb = nbformat.read(f, as_version=4)
        nbformat.validator.normalize(nb, version=4, version_minor=5)
        current_major = nb.get("nbformat")
        if not current_major or int(current_major) < 4:
            nb["nbformat"] = 4
        current_minor = nb.get("nbformat_minor")
        if not current_minor or int(current_minor) < 5:
            nb["nbformat_minor"] = 5

        with open(file.local_file_path(), "w", encoding="utf-8") as f:
            nbformat.write(nb, f)


def deploy_dir(repo: str, notebook_dir: str, normalize_source: bool) -> None:
    """
    Process each notebook in the specified directory, stamp it with Git info,
    and upload it to S3.
    """
    cfg = get_config("nbdeploy")
    temp_dir = cfg.get("temp_dir")
    if temp_dir is None:
        raise ValueError("nbdeploy config has no value for temp_dir")
    preserve_temp = cfg.get("preserve_temp", False)
    bucket = cfg.get("bucket")
    if bucket is None:
        raise ValueError("nbdeploy config has no value for bucket")
    print(f"Deploying notebooks from {repo}/{notebook_dir} to S3 bucket {bucket}")

    # get files to stamp
    notebook_files = list_files(repo, notebook_dir)
    if normalize_source:
        normalize(notebook_files)
    for nbfile in notebook_files:
        print(f"Stamping and uploading notebook: {nbfile.local_file_path()}.")
        stamped_path = nbfile.temp_file_path(temp_dir)
        if not os.path.exists(os.path.dirname(stamped_path)):
            os.makedirs(os.path.dirname(stamped_path), exist_ok=True)
        stamp_notebook(nbfile.local_file_path(), stamped_path)
        put_file(local_path=stamped_path, bucket=bucket, file_name=nbfile.s3_file_path())
        if not preserve_temp:
            os.remove(stamped_path)  # Clean up local copy after upload


def deploy_repo(repo_name: str, normalize_source: bool) -> None:
    """
    Deploy all notebooks in the specified repo, timestamping and tagging them with git revision.
    """
    # find repo
    repo_dir = find_dir(repo_name)
    if not repo_dir:
        print(f"Error: git repo '{repo_name}' not found")
        return
    os.chdir(repo_dir)
    repo_cfg = get_repo_config(repo_name)

    for dir in repo_cfg.get("directories", []):
        notebooks_dir = os.path.join(repo_dir, dir)
        if not os.path.exists(notebooks_dir):
            print(f"Error: notebooks dir '{notebooks_dir}' not found in {repo_dir}")
            continue
        deploy_dir(repo=repo_name, notebook_dir=dir, normalize_source=normalize_source)


def deploy_notebooks(repo: str = "all", normalize_source: bool = False) -> None:
    """Deploy notebooks to S3 from configured repo(s), timestamping and tagging them with git revision."""
    start_dir = os.getcwd()

    cfg = get_config("nbdeploy")
    temp_dir = cfg.get("temp_dir")
    if temp_dir is None:
        raise ValueError("nbdeploy config has no value for temp_dir")
    create_temp_dir = not os.path.exists(temp_dir)
    preserve_temp = bool(cfg.get("preserve_temp", False))
    if create_temp_dir:
        os.makedirs(temp_dir, exist_ok=True)

    if repo != "all":
        deploy_repo(repo_name=repo, normalize_source=normalize_source)
    else:
        for r in cfg.get("repos", []):
            deploy_repo(repo_name=r["name"], normalize_source=normalize_source)

    # cleanup
    os.chdir(start_dir)
    if create_temp_dir and not preserve_temp:
        for _, _, files in os.walk(temp_dir):
            if files:
                raise FileExistsError(f"{temp_dir} is not empty: {files}")
        print(f"removing temp dir: {temp_dir}")
        shutil.rmtree(temp_dir)
    print("Deployment complete!")




from pathlib import Path
import sys
import git
import yaml

import freeds.setup.utils as utils
from freeds.utils.root_config import RootConfig
import freeds.utils.log as log

logger = log.setup_logging(__name__)

def prompt_continue(user_dir: Path, this_dir: Path) -> bool:
    return utils.prompt_yesno(
        description=(
            f"Current folder: ({Path.cwd()})\n"
            f"Some repos will be cloned to this folder and we'll setup the config dir here.\n"
            f"One config file '.freeds' will be created in your user directory: {str(user_dir)}.\n"
            f"No other changes will be made outside this current folder unless you configure it."
        ),
        question="Shall we proceed",
    )


def prompt_keep_existing_config_file(file: Path) -> bool:
    if not file.exists():
        return True
    return


def setup_root_dir() -> bool:
    utils.log_header("Setting up the FreeDS root directory", "-")
    # this is the root folder, we're cloning repos and setting up lgos configs etc in this folder.
    root_path = Path.cwd()

    if not prompt_continue(user_dir=Path.home(), this_dir=root_path):
        return False

    git_repos = {
        "freeds-config": "https://github.com/jens-koster/freeds-config.git",
        "the-free-data-stack": "https://github.com/jens-koster/the-free-data-stack.git",
        "freeds-lab-databrickish": "https://github.com/jens-koster/freeds-lab-databrickish.git",
    }
    logger.info("Cloning git repos...")
    for name, url in git_repos.items():
        if (root_path / name).exists():
            logger.info(f"✅ Repo {name} already exists, skipping.")
            continue
        logger.info(f"Cloning repo: {url} into {name}...")
        logger.info(f"✅ Repo {name} cloned.")
        git.Repo.clone_from(url, name)

    root_config = RootConfig()

    if root_config.is_loaded:
        if not utils.prompt_yesno(
            description=f"The freeds config file already exists: {str(root_config.freeds_file_path())}.", question="Keep the existing config"
        ):
            root_config.set_default(root_path=root_path)

    logger.info(f"✅ Loaded config file: {root_config.freeds_file_path()}")
    logger.info(f"✅ Freeds root path: {root_config.root_path}")
    logger.info(f"✅ Config path (normally the config repo): {root_config.configs_path}")
    logger.info(f"✅ Locals (secrets) path: {root_config.locals_path}")

    logger.info("Creating directory structure")
    tfds_repo_root = root_path / "the-free-data-stack"
    data_root = root_path / "data"
    plugins_root = root_path / "plugins"

    paths = [
        plugins_root,
        plugins_root / "airflow",
        plugins_root / "airflow/logs",
        plugins_root / "postgres",
        plugins_root / "spark",
        data_root / "minio",
        data_root / "spark",
        data_root / "postgres",
        data_root / "local-pypi",
    ]

    for path in paths:
        logger.info(f"Creating {path}.")
        path.mkdir(parents=True, exist_ok=True)

    # Airflow
    airflow_symlink_root = plugins_root / "airflow"
    airflow_target_root = tfds_repo_root / "airflow"
    utils.relink(symlink=airflow_symlink_root / "dags", target=airflow_target_root / "dags")
    utils.relink(symlink=airflow_symlink_root / "config", target=airflow_target_root / "config")
    utils.relink(symlink=airflow_symlink_root / "plugins", target=airflow_target_root / "plugins")

    # postgres
    utils.relink(symlink=plugins_root / "postgres" / "init", target=tfds_repo_root / "postgres" / "init")

    # spark
    spark_root = plugins_root / "spark"
    utils.relink(symlink=spark_root / "conf", target=tfds_repo_root / "spark" / "conf")
    utils.relink(symlink=spark_root / "jars", target=tfds_repo_root / "spark" / "jars")

    logger.info(f"Setting up config dir for secrets: {root_config.locals_path}")
    # configs stay in the repo dir and are normally not changed.
    # we just copy the secrets files to locals which.
    root_config.locals_path.mkdir(parents=True, exist_ok=True)
    local_files = ["s3.yaml", "minio.yaml", "airflow.yaml", "postgres.yaml"]
    for file in local_files:
        utils.soft_copy(
            source=root_config.configs_path / file,
            target=root_config.locals_path / file
        )
    utils.log_header(title="🟢 Directory setup completed successfully 🌟", char=" ")
    return True

if __name__ == "__main__":
    setup_root_dir()



from freeds.config import get_env

def env()->None:
    """Print the export statements for freeds env values, allowing you to set the env values in a sceipt; `source <(freeds env)`. """
    for key, value in get_env().items():
        print(f'export {key}="{value}"')



"""check that all the freeds directories are in place."""

from typing import List

from freeds.utils import RootConfig
from freeds.selfcheck.check_classes import (
    AllGoodCheckResult,
    CheckList,
    CheckResult,
    PluginCheckResult,
)
from freeds.selfcheck.plugin_classes import get_plugins


def docker_compose_exists_check() -> list[CheckResult]:
    """Check that all docker compose files that should exist actually do."""
    result: list[CheckResult] = []
    for p in get_plugins():
        if not (p.path / "docker-compose.yaml").exists():
            result.append(
                PluginCheckResult(
                    passed=False,
                    plugin_name=p.name,
                    message=f"Plugin {p.name} in {p.path} does not have a docker-compose.yaml file.",
                )
            )
    if len(result) == 0:
        result.append(AllGoodCheckResult("All plugins have a docker-compose.yaml file."))
    return result


def readme_exists_check() -> list[CheckResult]:
    """Check that all plugins have a readme file."""
    result: list[CheckResult] = []
    for p in get_plugins():
        if not (p.path / "README.md").exists():
            result.append(
                PluginCheckResult(
                    passed=False,
                    plugin_name=p.name,
                    message=f"Plugin {p.name} in {p.path} does not have a README.md file.",
                )
            )
    if len(result) == 0:
        result.append(AllGoodCheckResult("All plugins have a README.md file."))
    return result


def check_directories_exist() -> List[CheckResult]:
    """
    check all the freeds folders are in place.
    """
    result: List[CheckResult] = []
    root = RootConfig().root_path
    expected_dirs = [
        ".",
        "local_configs",
        "freeds-config",
        "freeds-config/configs",
        "the-free-data-stack",
        "plugins/airflow",
        "plugins/airflow/config",
        "plugins/airflow/plugins",
        "plugins/airflow/dags",
        "plugins/spark",
        "plugins/spark/jars",
        "plugins/spark/conf",
        "plugins/postgres",
        "plugins/postgres/init",
        "logs",
        "data",
        "data/minio",
        "data/spark",
        "data/local-pypi",
    ]
    for dir_name in expected_dirs:
        dir_path = root / dir_name
        if not dir_path.exists():
            result.append(CheckResult(passed=False, message=f"Directory {dir_path} does not exist."))
    if len(result) == 0:
        result.append(AllGoodCheckResult(message=f"All expected directories are present in {root}."))
    return result


def checks() -> CheckList:
    """Get all checks related to web ui:s."""
    checklst = CheckList(area=__name__)

    checklst.add(
        name="Check freeds directories",
        description="Check that freeds directories and loinks exist, like config, secrets etc.",
        method=check_directories_exist,
    )
    checklst.add(
        name="Missing Docker Compose files",
        description="Check that all plugins have a docker-compose.yaml file.",
        method=docker_compose_exists_check,
    )
    checklst.add(
        name="Missing README.md files",
        description="Check that all plugins have a documentaiton file REAME.md.",
        method=readme_exists_check,
    )
    return checklst
